import numpy as np
from pystrict import strict
import PyMPDATA_examples.Arabas_and_Farhat_2020.Black_Scholes_1973 as BS73
import PyMPDATA_examples.Arabas_and_Farhat_2020.Bjerksund_and_Stensland_1993 as BS93


@strict
class Settings:
    amer = True
    l2_opt = 2.05
    S_min = .05
    S_max = 500
    K = 100
    r = .08
    sigma = .2
    n_iters = 2

    def __init__(self, *, T: float, C_opt: float, S0: float):
        self.T = T
        self.C_opt = C_opt
        self.S0 = S0
        self.S_match = S0

    def payoff(self, S: np.ndarray):
        return np.exp(-self.r * self.T) * (np.maximum(0, self.K - S))

    def analytical_solution(self, S: [np.ndarray, float], amer=True):
        if not amer:
            return BS73.p_euro(S, K=self.K, T=self.T, r=self.r, b=self.r, sgma=self.sigma)
        return BS93.p_amer(S, K=self.K, T=self.T, r=self.r, b=self.r, sgma=self.sigma)



import numpy as np
import numba
from pystrict import strict


grid = (100, 100)

dt = .1
dx = 1
dy = 1
omega = .1
h = 4.
h0 = 1

r = 15. * dx
x0 = 50 * dx
y0 = 75 * dy
xc = .5 * grid[0] * dx
yc = .5 * grid[1] * dy


@strict
class Settings:
    def __init__(self, n_rotations: int = 6):
        self.n_rotations = n_rotations

    @property
    def dt(self):
        return dt

    @property
    def nt(self):
        return int(628 * self.n_rotations)

    @property
    def size(self):
        return self.xrange[1], self.yrange[1]

    @property
    def xrange(self):
        return 0, grid[0] * dx

    @property
    def yrange(self):
        return 0, grid[1] * dy

    @property
    def grid(self):
        return grid

    @staticmethod
    @numba.njit()
    def pdf(x, y):
        tmp = (x-x0)**2 + (y-y0)**2
        return h0 + np.where(
            # if
            tmp - r**2 <= 0,
            # then
            h - np.sqrt(tmp / (r/h)**2),
            # else
            0.
        )

    @staticmethod
    def stream_function(xX, yY):
        x = xX * grid[0] * dx
        y = yY * grid[1] * dy
        return 1 / 2 * omega * ((x - xc)**2 + (y - yc)**2)



import numpy as np
from PyMPDATA import Options, Solver, Stepper, ScalarField, VectorField
from PyMPDATA.boundary_conditions import Periodic
from PyMPDATA_examples.utils.discretisation import from_cdf_1d
from PyMPDATA_examples.Smolarkiewicz_2006_Figs_3_4_10_11_12.settings import Settings


class Simulation:
    def __init__(self, settings: Settings, options: Options):

        _, state = from_cdf_1d(settings.cdf, settings.x_min, settings.x_max, settings.nx)

        self.stepper = Solver(
            stepper=Stepper(options=options, n_dims=len(state.shape), non_unit_g_factor=False),
            advectee=ScalarField(
                state.astype(options.dtype),
                halo=options.n_halo,
                boundary_conditions=(Periodic(),)
            ),
            advector=VectorField(
                (np.full(state.shape[0] + 1, settings.C, dtype=options.dtype),),
                halo=options.n_halo,
                boundary_conditions=(Periodic(),)
            )
        )
        self.nt = settings.nt

    @property
    def state(self):
        return self.stepper.advectee.get().copy()

    def run(self):
        self.stepper.advance(self.nt)



import numpy as np
import pint
from scipy import optimize
from scipy import integrate
from pystrict import strict
from PyMPDATA_examples.Olesik_et_al_2022 import equilibrium_drop_growth, East_and_Marshall_1954

default_nr = 64
default_GC_max = .5
default_mixing_ratios_g_kg = np.array([1, 2, 4, 6, 8, 10])
default_opt_set = {
    'a': {'n_iters': 1},
    'b': {'n_iters': 2},
    'c': {'n_iters': 2, 'infinite_gauge': True},
    'd': {'n_iters': 2, 'infinite_gauge': True, 'nonoscillatory': True},
    'e': {'n_iters': 2, 'DPDC': True, 'infinite_gauge': True, 'nonoscillatory': True},
    'f': {'n_iters': 3, 'third_order_terms': True},
    'g': {'n_iters': 3},
    'h': {'n_iters': 3, 'third_order_terms': True, 'infinite_gauge': True, 'nonoscillatory': True},
}
colors = ['red', 'blue', 'crimson', 'orange', 'olive', 'navy', 'green', 'blueviolet']
colors = {key: colors.pop(0) for key in default_opt_set}


def option_string(opts):
    str_repl = [["'n_iters': 1", "upwind"],
                ["'n_iters': 2", "MPDATA 2 iterations"],
                ["'n_iters': 3", "MPDATA 3 iterations"],
                ["'", ""],
                [": True", ""],
                ["_", " "],
                ["{", ""], ["}", ""], [",", ""], ["nonoscillatory", "non-oscillatory"]]
    for repl in str_repl:
        opts = opts.replace(repl[0], repl[1])
    return opts


si = pint.UnitRegistry()
ksi_1 = 100 * si.micrometre ** 2 / si.second


# based on Fig. 3 from East 1957
@strict
class Settings:
    def __init__(self, nr: int = default_nr,
                 mixing_ratios_g_kg: np.ndarray = default_mixing_ratios_g_kg):
        self.si = si
        self.nr = nr
        self.r_min = 1 * si.micrometre
        self.r_max = 25 * si.micrometre
        self.rho_w = 1 * si.kilogram / si.decimetre ** 3
        self.rho_a = 1 * si.kilogram / si.metre ** 3
        self.mixing_ratios = mixing_ratios_g_kg * si.gram / si.kilogram
        S = 1.00075
        self.drdt = equilibrium_drop_growth.DrDt(ksi_1, S)
        self.size_distribution = East_and_Marshall_1954.SizeDistribution(si)

        self.C = (1 * si.gram / si.kilogram) / self.mixing_ratio(self.size_distribution.pdf)
        np.testing.assert_approx_equal(self.C.to(si.dimensionless).magnitude, 1, 4)
        self.out_times = self.find_out_steps()

    def find_out_steps(self):
        out_steps = []
        for mr in self.mixing_ratios:
            t_unit = self.si.second
            def findroot(ti, mr=mr, t_unit=t_unit):
                return (mr - self.mixing_ratio(
                    equilibrium_drop_growth.PdfEvolver(self.pdf, self.drdt, ti * t_unit))).magnitude
            t = optimize.brentq(findroot, 0, (1 * self.si.hour).to(t_unit).magnitude)
            out_steps.append(t)
        return out_steps

    def mixing_ratio(self, pdf):
        xunit = self.si.micrometre
        yunit = 1 / self.si.micrometre / self.si.centimetre ** 3

        r_min = .1 * self.si.um
        while not np.isfinite(pdf(r_min).magnitude):
            r_min *= 1.01

        def pdfarg(r_nounit):
            r = r_nounit * xunit
            result = pdf(r) * r ** 3
            return result.to(yunit * xunit ** 3).magnitude

        I = integrate.quad(pdfarg,
                           r_min.to(xunit).magnitude,
                           np.inf
                           )[0] * yunit * xunit ** 4
        return (I * 4 / 3 * np.pi * self.rho_w / self.rho_a).to(self.si.gram / self.si.kilogram)

    def pdf(self, r):
        return self.C * self.size_distribution.pdf(r)



import pint
from scipy import integrate
from matplotlib import pyplot
import numpy as np
import pytest
from PyMPDATA_examples.Olesik_et_al_2022.East_and_Marshall_1954 import SizeDistribution
from PyMPDATA_examples.utils.discretisation import discretised_analytical_solution
from PyMPDATA_examples.Olesik_et_al_2022.coordinates import x_id, x_log_of_pn, x_p2


def diff(x):
    return np.diff(x.magnitude) * x.units


@pytest.mark.parametrize(
    "grid", [x_id(), x_log_of_pn(r0=1), x_p2()]
)
@pytest.mark.parametrize(
    "coord", [x_id(), x_log_of_pn(r0=1), x_p2()]
)
def test_size_distribution(grid, coord, plot=False):
    # Arrange
    si = pint.UnitRegistry()
    sd = SizeDistribution(si)
    n_unit = si.centimetres ** -3 / si.micrometre
    r_unit = si.micrometre

    # Act
    x = grid.x(np.linspace(1, 18, 100)) * r_unit
    dx_dr = coord.dx_dr
    numpdfx = x[1:] - diff(x) / 2

    def pdf_t(r):
        return sd.pdf(r * r_unit).to(n_unit).magnitude / dx_dr(r * r_unit).magnitude

    numpdfy = discretised_analytical_solution(rh=x.magnitude, pdf_t=pdf_t) * n_unit

    # Plot
    if plot:
        # Fig 3 from East & Marshall 1954
        si.setup_matplotlib()
        pyplot.plot(numpdfx, numpdfy, label='cdf')
        pyplot.plot(numpdfx, pdf_t(numpdfx.magnitude) * n_unit, label='pdf', linestyle='--')
        pyplot.legend()
        pyplot.gca().yaxis.set_units(1 / si.centimetre ** 3 / si.micrometre)
        pyplot.show()

    # Assert
    totalpdf = np.sum(numpdfy * (diff(x)))
    integratedpdf, _ = integrate.quad(pdf_t, x[0].magnitude, x[-1].magnitude)
    print(totalpdf, integratedpdf)
    np.testing.assert_array_almost_equal(totalpdf.magnitude, integratedpdf)  # pylint: disable=no-member


def test_quad_vs_midpoint():
    # Arrange
    x = np.linspace(0, np.pi, 3)
    a = discretised_analytical_solution(x, np.sin, midpoint_value=True, r=x[:-1] + np.diff(x)/2)
    b = discretised_analytical_solution(x, np.sin, midpoint_value=False)
    test_val = np.abs((a - b)/a)
    # Assert
    assert (test_val > .05).all()
    assert (test_val < .1).all()



# pylint: disable=wrong-import-position
# https://bugs.python.org/issue37373
import sys
if sys.platform == 'win32' and sys.version_info[:2] >= (3, 7):
    import asyncio
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

import pathlib
import re
import os
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor
import pytest


# https://stackoverflow.com/questions/7012921/recursive-grep-using-python
def findfiles(path, regex):
    regObj = re.compile(regex)
    res = []
    for root, _, fnames in os.walk(path):
        for fname in fnames:
            if regObj.match(fname):
                res.append(os.path.join(root, fname))
    return res


@pytest.fixture(params=findfiles(pathlib.Path(__file__).parent.parent.absolute(), r'.*\.(ipynb)$'))
def notebook_filename(request):
    return request.param


def test_run_notebooks(notebook_filename, tmp_path):  # pylint: disable=redefined-outer-name
    with open(notebook_filename, encoding="utf8") as f:
        nb = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=15*60, kernel_name='python3')
        ep.preprocess(nb, {'metadata': {'path': tmp_path}})



import numpy as np


def L2(numerical, analytical, nt):
    assert numerical.shape == analytical.shape
    N = analytical.size
    err2 = np.log(
        np.sqrt(
            sum(pow(numerical - analytical, 2)) / nt / N
        )
    ) / np.log(2)
    return err2


def Smolarkiewicz_Grabowski_1990_eq21(numerical, analytical, T):
    assert numerical.shape == analytical.shape
    NX = analytical.size
    err = np.sqrt(
            sum(pow(numerical - analytical, 2)) / NX
        ) / T
    return err


def modified_Smolarkiewicz_Rasch_r0(numerical, analytical, T, g_factor):
    NX = analytical.size
    err = np.sqrt(sum(pow(numerical - analytical, 2) * g_factor) / NX) / np.amax(analytical) / T
    return err



from functools import lru_cache
import numpy as np
import numba
from PyMPDATA.impl.enumerations import (SIGN_RIGHT,
                                        ARG_DATA, OUTER, ARG_FOCUS,
                                        META_AND_DATA_DATA, META_AND_DATA_META)


@lru_cache()
def _make_scalar(value, _, halo, __, ___, dtype, jit_flags):
    @numba.njit(**jit_flags)
    def impl(psi, ____, sign):
        if sign == SIGN_RIGHT:
            return 0
        z = psi[ARG_FOCUS][OUTER]
        activated = np.sum(psi[ARG_DATA][z:z+1, halo:-halo])
        # assert activated < value
        result = max(0, value - activated)
        return result

    if dtype == complex:
        @numba.njit(**jit_flags)
        def fill_halos_scalar(psi, n, sign):
            return complex(
                impl((psi[META_AND_DATA_META], psi[META_AND_DATA_DATA].real), n, sign),
                impl((psi[META_AND_DATA_META], psi[META_AND_DATA_DATA].imag), n, sign)
            )
    else:
        @numba.njit(**jit_flags)
        def fill_halos_scalar(psi, n, sign):
            return impl(psi, n, sign)
    return fill_halos_scalar


# pylint: disable=too-few-public-methods
class DropletActivation:
    def __init__(self, value, dr, dz):
        self._value = value / dz / dr
        self.dz = dz
        self.dr = dr

    def make_scalar(self, _at, _halo, dtype, jit_flags):
        return _make_scalar(self._value, _at, _halo, self.dr, self.dz, dtype, jit_flags)



from collections import namedtuple
from copy import deepcopy
import numpy as np
from joblib import Parallel, parallel_backend, delayed
from PyMPDATA import Options
from PyMPDATA_examples.Olesik_et_al_2022.simulation import Simulation
from PyMPDATA_examples.Olesik_et_al_2022.equilibrium_drop_growth import PdfEvolver
from PyMPDATA_examples.utils.error_norms import L2
from PyMPDATA_examples.utils.discretisation import discretised_analytical_solution
from PyMPDATA_examples.Olesik_et_al_2022.coordinates import x_id, x_log_of_pn, x_p2
from PyMPDATA_examples.Olesik_et_al_2022.settings import Settings, default_mixing_ratios_g_kg


def analysis(settings, grid_layout, psi_coord, options_dict, GC_max):
    options_str = str(options_dict)
    options = Options(**options_dict)
    simulation = Simulation(settings, grid_layout, psi_coord, options, GC_max)
    result = {"n": [], "n_analytical": [], "error_norm_L2": [], "wall_time":[]}
    last_step = 0
    for n_steps in simulation.out_steps:
        steps = n_steps - last_step
        wall_time = simulation.step(steps) if steps > 0 else 0
        last_step += steps
        result['n'].append(simulation.n_of_r.copy())
        result['wall_time'].append(wall_time)
    result['r'] = simulation.r.copy()
    result['rh'] = simulation.rh.copy()
    result['dx'] = simulation.dx
    return Result(
        grid_layout_str=grid_layout.__class__.__name__,
        option_str=options_str,
        result=result,
        out_steps=simulation.out_steps,
        dt=simulation.dt
    )


def compute_figure_data(*, nr, GC_max, psi_coord=x_id(),
                        grid_layouts=(x_id(), x_p2(), x_log_of_pn(r0=1, n=1)),
                        opt_set=({'n_iters': 1},),
                        mixing_ratios_g_kg=default_mixing_ratios_g_kg
                        ):
    settings = Settings(nr=nr, mixing_ratios_g_kg=mixing_ratios_g_kg)
    with parallel_backend('threading', n_jobs=-2):
        results = Parallel(verbose=10)(
            delayed(analysis)(settings, grid_layout, psi_coord, options, GC_max)
            for grid_layout in grid_layouts
            for options in deepcopy(opt_set)
        )

    cases = {}
    for result in results:
        case = Case(result)
        if case.grid_layout_str not in cases:
            cases[case.grid_layout_str] = case


    output = {}
    for coord, case in cases.items():
        output[coord] = {"numerical": {}, "wall_time": {}}
        for result in results:
            if coord == result.grid_layout_str:
                opts = result.option_str
                data = result.result
                rh = data.pop("rh")
                r = data.pop('r')
                dx = data.pop('dx')

                if 'grid' not in output[coord]:
                    output[coord]["grid"] = {
                        'rh': rh, 'r': r, 'dx': dx, 'dt': case.dt, 'out_steps': case.out_steps
                    }
                output[coord]["numerical"][opts] = data['n']

    for coord, case in cases.items():
        analytical = []
        for t in [case.dt * nt for nt in case.out_steps]:
            pdf_t = PdfEvolver(settings.pdf, settings.drdt, t)
            rh = output[coord]["grid"]['rh']
            analytical.append(discretised_analytical_solution(
                rh.magnitude,
                lambda r, pdf_t=pdf_t, rh=rh: pdf_t(r * rh.units).magnitude,
                midpoint_value=True,
                r=output[coord]["grid"]['r'].magnitude
            ) * pdf_t(rh[0]).units)
        output[coord]["analytical"] = analytical

    for coord, case in cases.items():
        error_L2 = {}
        analytical = output[coord]["analytical"]
        for opts in output[coord]["numerical"]:
            numerical = output[coord]["numerical"][opts]
            error_L2[opts] = L2(
                numerical[-1].magnitude,
                analytical[-1].magnitude,
                case.out_steps[-1]
            )
        output[coord]["error_L2"] = error_L2

    for coord, case in cases.items():
        for result in results:
            data = result.result
            for opts in output[coord]["numerical"]:
                output[coord]["wall_time"][opts] = data["wall_time"]

    return output, settings


class Result(namedtuple("Result", ('dt', 'out_steps', 'grid_layout_str', 'option_str', 'result'))):
    __slots__ = ()


def Case(result: Result):
    return namedtuple("Case", ('dt', 'out_steps', 'grid_layout_str'))(
        dt=result.dt,
        out_steps=result.out_steps,
        grid_layout_str=result.grid_layout_str
    )


def rel_disp(r, psi, psi_coord):
    mom0 = 0
    mom1 = 0
    mom2 = 0
    for i, psi_i in enumerate(psi):
        dp_i = (
            psi_coord.moment_of_r_integral(psi_coord.x(r[i+1]), 0) -
            psi_coord.moment_of_r_integral(psi_coord.x(r[i]), 0)
        )
        A_i = (
            psi_coord.moment_of_r_integral(psi_coord.x(r[i+1]), 1) -
            psi_coord.moment_of_r_integral(psi_coord.x(r[i]), 1)
        )
        B_i = (
            psi_coord.moment_of_r_integral(psi_coord.x(r[i+1]), 2) -
            psi_coord.moment_of_r_integral(psi_coord.x(r[i]), 2)
        )
        bin_mom0 = psi_i * dp_i
        bin_mom1 = psi_i * A_i
        bin_mom2 = psi_i * B_i
        mom0 += bin_mom0
        mom1 += bin_mom1
        mom2 += bin_mom2
    mu = mom1 / mom0
    std = np.sqrt(mom2/mom0 - mu**2)
    d = std/mu
    return d


def third_moment(r, psi, psi_coord, normalised=True):
    mom0 = 0
    mom3 = 0
    for i, psi_i in enumerate(psi):
        dp_i = (psi_coord.moment_of_r_integral(psi_coord.x(r[i+1]), 0) -
                psi_coord.moment_of_r_integral(psi_coord.x(r[i]), 0))
        integral_i = (
                psi_coord.moment_of_r_integral(psi_coord.x(r[i+1]), 3) -
                psi_coord.moment_of_r_integral(psi_coord.x(r[i]), 3))
        bin_mom0 = psi_i * dp_i
        bin_mom3 = psi_i * integral_i
        mom0 += bin_mom0
        mom3 += bin_mom3
    if normalised:
        mom3 /= mom0
    return mom3



from joblib import Parallel, delayed
from PyMPDATA import Options
from PyMPDATA_examples.Molenkamp_test_as_in_Jaruga_et_al_2015_Fig_12.simulation import Simulation
from PyMPDATA_examples.Molenkamp_test_as_in_Jaruga_et_al_2015_Fig_12.settings import Settings


options = {
    'upwind': Options(n_iters=1),
    '2+fct': Options(n_iters=2, nonoscillatory=True),
    '3+fct+tot': Options(n_iters=3, nonoscillatory=True, third_order_terms=True),
    '2+fct+iga': Options(n_iters=2, nonoscillatory=True, infinite_gauge=True)
}


def compute_panel(panel):
    settings = Settings(n_rotations=6)
    simulation = Simulation(settings, options[panel])
    if panel == 'upwind':
        return simulation.state
    simulation.run()
    return simulation.state


def fig_12_data():
    data = Parallel(n_jobs=-2)(
        delayed(compute_panel)(panel)
        for panel in ['upwind', '2+fct', '3+fct+tot', '2+fct+iga']
    )
    return data



import numpy as np


# pylint: disable=too-few-public-methods
class SizeDistribution:
    def __init__(self, si):
        self.si = si
        self.n0 = 465 / si.centimetre**3  # adjusted wrt original 700 to match mixing ratio of 1g/kg
        self.kappa = 22

    def pdf(self, r):
        return (
                (self.n0 * self.si.micrometre) / r *
                np.exp(-self.kappa * (np.log10(r / (7 * self.si.micrometre)) ** 2)) *
                (1 / self.si.micrometre))



from pystrict import strict
from PyMPDATA import Options


@strict
class Settings:
    def __init__(self):
        self.dt = .01
        self.dx = .05
        self.dy = .05
        self.nx = 401
        self.ny = 401
        self.eps = 1e-7
        self.lx0 = 2
        self.ly0 = 1
        self.options = Options(
            nonoscillatory=True,
            infinite_gauge=True
        )

    @property
    def nt(self):
        return int(7 / self.dt)

    @property
    def outfreq(self):
        return int(1 / self.dt)



import numpy as np
from PyMPDATA import ScalarField, VectorField, Solver, Stepper
from PyMPDATA.boundary_conditions import Constant


class Simulation:
    def __init__(self, settings, options, static=True, n_threads=None):
        bcs = tuple(Constant(0) for _ in settings.grid)

        advector = VectorField(
            data=tuple(comp.astype(options.dtype) for comp in settings.advector),
            halo=options.n_halo,
            boundary_conditions=bcs
        )

        advectee = ScalarField(
            data=np.asarray(settings.advectee, dtype=options.dtype),
            halo=options.n_halo,
            boundary_conditions=bcs
        )

        args = {'grid': settings.grid} if static else {'n_dims': len(settings.grid)}
        if n_threads is not None:
            args['n_threads'] = n_threads
        stepper = Stepper(options=options, **args)
        self.solver = Solver(stepper=stepper, advectee=advectee, advector=advector)

    def run(self, nt):
        return self.solver.advance(nt)

    @property
    def advectee(self):
        return self.solver.advectee



import math
import numpy as np
from PyMPDATA import ScalarField, VectorField, Stepper, Solver, Options
from PyMPDATA.boundary_conditions import Constant, Extrapolated
from PyMPDATA_examples.utils.discretisation import discretised_analytical_solution


class Simulation:
    @staticmethod
    def make_condensational_growth_solver(nr, r_min, r_max,
                                          GC_max, grid_layout, psi_coord,
                                          pdf_of_r, drdt_of_r, opts: Options):
        # psi = psi(p)
        dp_dr = psi_coord.dx_dr
        dx_dr = grid_layout.dx_dr

        xh, dx = np.linspace(
            grid_layout.x(r_min),
            grid_layout.x(r_max),
            nr + 1,
            retstep=True
        )
        rh = grid_layout.r(xh)

        x = np.linspace(
            xh[0] + dx / 2,
            xh[-1] - dx / 2,
            nr
        )
        r = grid_layout.r(x)

        def pdf_of_r_over_psi(r):
            return pdf_of_r(r) / psi_coord.dx_dr(r)

        psi = discretised_analytical_solution(rh, pdf_of_r_over_psi, midpoint_value=True, r=r)

        dp_dt = drdt_of_r(rh) * dp_dr(rh)
        G = dp_dr(r) / dx_dr(r)

        # C = dr_dt * dt / dr
        # GC = dp_dr / dx_dr * dr_dt * dt / dr =
        #        \       \_____ / _..____/
        #         \_____.._____/    \_ dt/dx
        #               |
        #             dp_dt

        dt = GC_max * dx / np.amax(dp_dt)
        GCh = dp_dt * dt / dx

        # CFL condition
        np.testing.assert_array_less(np.abs(GCh), 1)

        g_factor = ScalarField(
            G.astype(dtype=opts.dtype),
            halo=opts.n_halo,
            boundary_conditions=(Extrapolated(),))
        state = ScalarField(
            psi.astype(dtype=opts.dtype),
            halo=opts.n_halo,
            boundary_conditions=(Constant(0),))
        GC_field = VectorField(
            [GCh.astype(dtype=opts.dtype)],
            halo=opts.n_halo,
            boundary_conditions=(Constant(0),))
        stepper = Stepper(
            options=opts,
            n_dims=1,
            non_unit_g_factor=True
        )
        return (
            Solver(stepper=stepper, g_factor=g_factor, advectee=state, advector=GC_field),
            r,
            rh,
            dx,
            dt,
            g_factor
        )


    @staticmethod
    def __mgn(quantity, unit):
        return quantity.to(unit).magnitude

    def __init__(self, settings, grid_layout, psi_coord, opts, GC_max):
        self.settings = settings
        self.psi_coord = psi_coord
        self.grid_layout = grid_layout

        # units of calculation
        self.__t_unit = self.settings.si.seconds
        self.__r_unit = self.settings.si.micrometre
        self.__p_unit = psi_coord.x(self.__r_unit)
        self.__n_of_r_unit = self.settings.si.centimetres ** -3 / self.settings.si.micrometre

        self.solver, self.__r, self.__rh, self.dx, dt, self._g_factor = \
            Simulation.make_condensational_growth_solver(
                self.settings.nr,
                self.__mgn(self.settings.r_min, self.__r_unit),
                self.__mgn(self.settings.r_max, self.__r_unit),
                GC_max,
                grid_layout,
                psi_coord,
                lambda r: self.__mgn(
                    self.settings.pdf(r * self.__r_unit), self.__n_of_r_unit),
                lambda r: self.__mgn(
                    self.settings.drdt(r * self.__r_unit), self.__r_unit / self.__t_unit),
                opts
            )

        self.out_steps = tuple(math.ceil(t/dt) for t in settings.out_times)
        self.dt = dt * self.__t_unit

    def step(self, nt):
        return self.solver.advance(nt)

    @property
    def r(self):
        return self.__r * self.__r_unit

    @property
    def rh(self):
        return self.__rh * self.__r_unit

    @property
    def n_of_r(self):
        psi = self.solver.advectee.get()
        n = psi * self.psi_coord.dx_dr(self.__r)
        return n * self.__n_of_r_unit

    @property
    def psi(self):
        psi_unit = self.__n_of_r_unit * self.__r_unit / self.__p_unit
        return self.solver.advectee.get() * psi_unit

    @property
    def g_factor(self):
        return self._g_factor.get()

    @property
    def dp_dr(self):
        return self.psi_coord.dx_dr(self.__r)



import numpy as np


class x_id:
    @staticmethod
    def x(r):
        return r

    @staticmethod
    def r(x):
        return x

    @staticmethod
    def dx_dr(r):
        return r**0

    @staticmethod
    def moment_of_r_integral(x, k):
        return 1 / (k + 1) * x ** (k + 1)


class x_p2:
    @staticmethod
    def x(r):
        return r**2

    @staticmethod
    def r(x):
        return np.sqrt(np.where(x < 0, 1e10, x))

    @staticmethod
    def dx_dr(r):
        return 2*r

    @staticmethod
    def moment_of_r_integral(x, k):
        return 2 / (k + 2) * x**((k+2)/2)


class x_p3:
    @staticmethod
    def x(r):
        return r**3

    @staticmethod
    def r(x):
        return np.power(x, 1/3)

    @staticmethod
    def dx_dr(r):
        return 3*r**2

    @staticmethod
    def moment_of_r_integral(x, k):
        return 3 / (k + 3) * x**((k+3)/3)


class x_log_of_pn:
    def __init__(self, r0=1, base=np.e, n=3):
        self.r0 = r0
        self.base = base
        self.n = n

    def x(self, r):
        return np.log(r**self.n / self.r0**self.n) / np.log(self.base)

    def r(self, x):
        return self.r0 * self.base**(x/self.n)

    def dx_dr(self, r):
        return self.n / r / np.log(self.base)

    def moment_of_r_integral(self, x, k):
        return self.r0**k * self.n / (k * np.log(self.base)) * self.base**(k / self.n * x)



import numpy as np
from PyMPDATA import ScalarField, VectorField, Solver, Stepper
from PyMPDATA.boundary_conditions import Constant
from PyMPDATA_examples.Jarecka_et_al_2015 import formulae


class Simulation:
    def __init__(self, settings):
        self.settings = settings
        s = settings

        halo = settings.options.n_halo
        grid = (s.nx, s.ny)
        bcs = [Constant(value=0)] * len(grid)

        self.advector = VectorField((
            np.zeros((s.nx + 1, s.ny)),
            np.zeros((s.nx, s.ny + 1))
        ), halo, bcs)

        xi, yi = np.indices(grid, dtype=float)
        xi -= (s.nx-1) / 2
        yi -= (s.ny-1) / 2
        x = xi * s.dx
        y = yi * s.dy
        h0 = formulae.amplitude(x, y, s.lx0, s.ly0)

        advectees = {
            'h': ScalarField(h0, halo, bcs),
            'uh': ScalarField(np.zeros(grid), halo, bcs),
            'vh': ScalarField(np.zeros(grid), halo, bcs)
        }

        stepper = Stepper(options=s.options, grid=grid)
        self.solvers = {
            k: Solver(stepper, v, self.advector)
            for k, v in advectees.items()
        }

    @staticmethod
    def interpolate(psi, axis):
        idx = (
            (slice(None, -1), slice(None, None)),
            (slice(None, None), slice(None, -1))
        )
        return np.diff(psi, axis=axis) / 2 + psi[idx[axis]]

    def run(self):
        s = self.settings
        grid_step = (s.dx, s.dy)
        idx = (
            (slice(1, -1), slice(None, None)),
            (slice(None, None), slice(1, -1))
        )
        output = []
        for it in range(s.nt + 1):
            if it != 0:
                h = self.solvers['h'].advectee.get()
                for xy, k in enumerate(('uh', 'vh')):
                    mask = h > s.eps
                    vel = np.where(mask, np.nan, 0)
                    np.divide(self.solvers[k].advectee.get(), h, where=mask, out=vel)
                    self.advector.get_component(xy)[idx[xy]] = \
                        self.interpolate(vel, axis=xy) * s.dt / grid_step[xy]
                self.solvers['h'].advance(1)
                assert h.ctypes.data == self.solvers['h'].advectee.get().ctypes.data
                for xy, k in enumerate(('uh', 'vh')):
                    psi = self.solvers[k].advectee.get()
                    psi[:] -= s.dt / 2 * h * np.gradient(h, grid_step[xy], axis=xy)
                    self.solvers[k].advance(1)
                    psi[:] -= s.dt / 2 * h * np.gradient(h, grid_step[xy], axis=xy)
            if it % s.outfreq == 0:
                output.append({
                    k: self.solvers[k].advectee.get().copy()
                    for k in self.solvers.keys()
                })
        return output



import numpy as np
import numba
from PyMPDATA.boundary_conditions import Extrapolated
from PyMPDATA import Options, Solver, Stepper, ScalarField, VectorField
from PyMPDATA_examples.Arabas_and_Farhat_2020.options import OPTIONS


class Simulation:
    @staticmethod
    def _factory(*,
                 options: Options,
                 advectee: np.ndarray,
                 advector: float,
                 boundary_conditions
                 ):
        stepper = Stepper(options=options, n_dims=len(advectee.shape), non_unit_g_factor=False)
        return Solver(
            stepper=stepper,
            advectee=ScalarField(
              advectee.astype(dtype=options.dtype),
              halo=options.n_halo,
              boundary_conditions=boundary_conditions
            ),
            advector=VectorField(
              (np.full(advectee.shape[0] + 1, advector, dtype=options.dtype),),
              halo=options.n_halo,
              boundary_conditions=boundary_conditions
)
        )

    def __init__(self, settings):
        self.settings = settings

        sigma2 = pow(settings.sigma, 2)
        dx_opt = abs(settings.C_opt / (.5 * sigma2 - settings.r) * settings.l2_opt * sigma2)
        dt_opt = pow(dx_opt, 2) / sigma2 / settings.l2_opt

        # adjusting dt so that nt is integer
        self.dt = settings.T
        self.nt = 0
        while self.dt > dt_opt:
            self.nt += 1
            self.dt = settings.T / self.nt

        # adjusting dx to match requested l^2
        dx = np.sqrt(settings.l2_opt * self.dt) * settings.sigma

        # calculating actual u number and lambda
        self.C = - (.5 * sigma2 - settings.r) * (-self.dt) / dx
        self.l2 = dx * dx / sigma2 / self.dt

        # adjusting nx and setting S_beg, S_end
        S_beg = settings.S_match
        self.nx = 1
        while S_beg > settings.S_min:
            self.nx += 1
            S_beg = np.exp(np.log(settings.S_match) - self.nx * dx)

        self.ix_match = self.nx

        S_end = settings.S_match
        while S_end < settings.S_max:
            self.nx += 1
            S_end = np.exp(np.log(S_beg) + (self.nx-1) * dx)

        # asset price
        self.S = np.exp(np.log(S_beg) + np.arange(self.nx) * dx)

        self.mu_coeff = (0.5 / self.l2,)
        self.solvers = {}
        self.solvers[1] = self._factory(
            advectee=settings.payoff(self.S),
            advector=self.C,
            options=Options(n_iters=1, non_zero_mu_coeff=True),
            boundary_conditions=(Extrapolated(),)
        )
        self.solvers[2] = self._factory(
            advectee=settings.payoff(self.S),
            advector=self.C,
            options=Options(**OPTIONS),
            boundary_conditions=(Extrapolated(),)
        )

    def run(self, n_iters: int):
        if self.settings.amer:
            psi = self.solvers[n_iters].advectee.data
            f_T = np.empty_like(psi)
            f_T[:] = psi[:] / np.exp(-self.settings.r * self.settings.T)
            T = self.settings.T
            r = self.settings.r
            dt = self.dt

            @numba.experimental.jitclass([])
            class PostStep:
                # pylint: disable=too-few-public-methods
                def __init__(self):
                    pass

                def __call__(self, psi, step):
                    t = T - (step + 1) * dt
                    psi += np.maximum(psi, f_T / np.exp(r * t)) - psi

            self.solvers[n_iters].advance(self.nt, self.mu_coeff, PostStep())
        else:
            self.solvers[n_iters].advance(self.nt, self.mu_coeff)

        return self.solvers[n_iters].advectee.get()

    def terminal_value(self):
        return self.solvers[1].advectee.get()



import numpy as np
import PyMPDATA_examples.Arabas_and_Farhat_2020.Black_Scholes_1973 as BS


def _phi(S: [np.ndarray, float], gamma: float, H: float, I: float, r: float,
         b: float, var: float, T: float):
    lmbd = (-r + gamma * b + .5 * gamma * (gamma - 1) * var) * T
    d = -(np.log(S / H) + (b + (gamma - .5) * var) * T) / np.sqrt(var * T)
    kappa = 2 * b / var + (2 * gamma - 1)
    return (np.exp(lmbd)
        * np.power(S, gamma)
        * (BS.N(d) - pow((I / S), kappa) * BS.N(d - 2 * np.log(I / S) / np.sqrt(var * T)))
    )


def c_amer(S: [np.ndarray, float], K: [float, np.ndarray], T: float, r: float,
           b: float, sgma: float):
    if b >= r:
        return BS.c_euro(S, K=K, T=T, r=r, b=b, sgma=sgma)

    var = sgma*sgma
    beta = (.5 - b/var) + np.sqrt(pow((b/var - .5), 2) + 2 * r/var)
    BInf = beta / (beta - 1) * K
    B0 = np.maximum(K, r / (r - b) * K)
    ht = -(b*T + 2 * sgma * np.sqrt(T)) * B0 / (BInf - B0)
    I = B0 + (BInf - B0) * (1 - np.exp(ht))
    alpha = (I-K) * pow(I, -beta)

    return np.where(
        S >= I,
        S - K,
        alpha * np.power(S, beta) + (
            - alpha * _phi(S, gamma=beta, H=I, I=I, r=r, b=b, var=var, T=T)
            +         _phi(S, gamma=1,    H=I, I=I, r=r, b=b, var=var, T=T)
            -         _phi(S, gamma=1,    H=K, I=I, r=r, b=b, var=var, T=T)
            -     K * _phi(S, gamma=0,    H=I, I=I, r=r, b=b, var=var, T=T)
            +     K * _phi(S, gamma=0,    H=K, I=I, r=r, b=b, var=var, T=T)
        )
    )


def p_amer(S: [np.ndarray, float], K: float, T: float, r: float, b: float, sgma: float):
    return c_amer(K, K=S, T=T, r=r - b, b=-b, sgma=sgma)



from pystrict import strict
import numpy as np


@strict
class Settings:
    def __init__(self, n: int, dt: float):
        self.grid = (n, n, n)
        self.dt = dt
        self.L = 100
        self.dx = self.L / n
        self.dy = self.dx
        self.dz = self.dx
        self.h = 4
        self.r = 15
        d = 25 / np.sqrt(3)
        self.x0 = 50 - d
        self.y0 = 50 + d
        self.z0 = 50 + d

        self.omega = 0.1
        self.xc = 50
        self.yc = 50
        self.zc = 50

    @property
    def advector(self):
        """ constant angular velocity rotational field """

        data = [None, None, None]
        for index, letter in enumerate(('x', 'y', 'z')):
            i, j, k = np.indices((g + (gi == index) for gi, g in enumerate(self.grid)))
            if letter == 'x':
                data[index] = (-((j+.5) * self.dy - self.yc) + ((k+.5) * self.dz - self.zc))\
                              / self.dx
            elif letter == 'y':
                data[index] = (+((i+.5) * self.dx - self.xc) - ((k+.5) * self.dz - self.zc))\
                              / self.dy
            elif letter == 'z':
                data[index] = (-((i+.5) * self.dx - self.xc) + ((j+.5) * self.dy - self.yc))\
                              / self.dz
            data[index] *= self.omega / np.sqrt(3) * self.dt
        return data

    @property
    def advectee(self):
        i, j, k = np.indices(self.grid)
        dist = (
            ((i+.5) * self.dx - self.x0) ** 2 +
            ((j+.5) * self.dy - self.y0) ** 2 +
            ((k+.5) * self.dz - self.z0) ** 2
        )
        return np.where(dist - pow(self.r, 2) <= 0, self.h, 0)



import numpy as np
from pystrict import strict


@strict
class Settings:
    nt = 1600
    dt = 1
    nx = 500
    C = 0.5
    x_min = -250
    x_max = 250

    def __init__(self, shape: str):
        if shape == "cosine":
            self.cdf = Settings.cdf_cosine
        elif shape == "rect":
            self.cdf = Settings.cdf_rect
        else:
            raise ValueError()

    @staticmethod
    def cdf_cosine(x):
        x_mid = -150
        f = 2/12
        amplitude = 2

        pdf = np.where(
            np.abs(x-x_mid) < 10,
            amplitude * np.cos(f*(x-x_mid)),
            0)
        return np.cumsum(pdf)

    @staticmethod
    def cdf_rect(x):
        x_mid = -150
        amplitude = 2
        offset = 2

        pdf = offset + np.where(
            np.abs(x-x_mid) <= 10,
            amplitude,
            0)
        return np.cumsum(pdf)



from matplotlib import pylab
import numpy as np
from PyMPDATA_examples.Jarecka_et_al_2015 import formulae


def plot_output(times, output, settings, return_data=True):
    lambdas_analytic = formulae.d2_el_lamb_lamb_t_evol(
        times=[0] + list(times),
        lamb_x0=settings.lx0,
        lamb_y0=settings.ly0
    )[1:]

    data = {}
    cuts = ('x', 'y')
    _, axs = pylab.subplots(
        nrows=len(times),
        ncols=len(cuts),
        figsize=(9, 9),
        constrained_layout=True
    )
    momenta = {'x': 'uh', 'y': 'vh'}
    for i_cut, cut in enumerate(cuts):
        if cut == 'x':
            n = settings.nx
            idx = (slice(None, None), slice(n // 2, n // 2 + 1))
            coord = settings.dx * (np.linspace(-n // 2, n // 2, n) + .5)
            x = coord
            y = 0
        else:
            n = settings.ny
            idx = (slice(n // 2, n // 2 + 1), slice(None, None))
            coord = settings.dy * (np.linspace(-n // 2, n // 2, n) + .5)
            x = 0
            y = coord
        for i_t, t in enumerate(times):
            key = f"cut={cut} t={t}"
            data[key] = {'coord': coord, 'x':x, 'y':y}
            datum = data[key]
            datum['h_analytic'] = formulae.amplitude(x, y, *lambdas_analytic[i_t, (0, 2)])
            where = 'mid'
            axs[i_t, i_cut].set_xlim(-8.5, 8.5)
            axs[i_t, i_cut].set_xticks((-5, 0, 5))
            axs[i_t, i_cut].step(datum['coord'], output[0]['h'][idx].squeeze(),
                                 color='black', label='h @ t=0', where=where)
            axs[i_t, i_cut].step(datum['coord'], output[t]['h'][idx].squeeze(),
                                 color='red', label='MPDATA: h', where=where)
            axs[i_t, i_cut].plot(datum['coord'], datum['h_analytic'],
                                 color='blue', linestyle=':', label='analytic: h')
            twin = axs[i_t, i_cut].twinx()
            datum['h_numeric'] = output[t]['h'][idx].squeeze()
            mask = datum['h_numeric'] > settings.eps
            datum['q_h_numeric'] = np.where(mask, np.nan, 0)
            np.divide(output[t][momenta[cut]][idx].squeeze(), datum['h_numeric'],
                      where=mask, out=datum['q_h_numeric'])
            twin.step(
                datum['coord'],
                datum['q_h_numeric'],
                color='red',
                linestyle='--',
                label='MPDATA: q/h',
                where=where
            )
            datum['q_h_analytic'] = np.where(
                datum['h_analytic'] > settings.eps,
                datum['coord'] / lambdas_analytic[i_t, 0 + 2 * i_cut] \
                    * lambdas_analytic[i_t, 1 + 2 * i_cut],
                0)
            twin.plot(
                datum['coord'],
                datum['q_h_analytic'],
                linestyle='-.',
                label='analytic: q/h',
                color='blue'
            )
            twin.set_ylabel('u' if cut == 'x' else 'v')
            twin.set_ylim(-1.2, 1.2)
            twin.set_yticks((-1, 0, 1))
            axs[i_t, i_cut].set_xlabel(cut)
            axs[i_t, i_cut].set_ylim(-.6, .6)
            axs[i_t, i_cut].set_yticks((-.5, 0, .5))
            axs[i_t, i_cut].set_ylabel('h')
            axs[i_t, i_cut].set_title(f"t={t}", y=1.0, pad=-14, x=0.075)
            axs[i_t, i_cut].grid()

    legend_props = {'frameon': False, 'fontsize': 'small'}
    axs[-1, -1].legend(loc='lower right', **legend_props)
    twin.legend(loc='lower center', **legend_props)
    if return_data:
        return data
    return None



import numpy as np
from scipy.special import erf  # pylint: disable=no-name-in-module


def N(x: float):
    return (1 + erf(x / np.sqrt(2))) / 2


def c_euro(S: np.ndarray, K: float, T: float, r: float, b: float, sgma: float):
    d1 = (np.log(S / K) + (b + sgma * sgma / 2) * T) / sgma / np.sqrt(T)
    d2 = d1 - sgma * np.sqrt(T)
    return S * np.exp(b - r) * N(d1) - K * np.exp(-r * T) * N(d2)


def p_euro(S: np.ndarray, K: float, T: float, r: float, b: float, sgma: float):
    d1 = (np.log(S / K) + (b + sgma * sgma / 2) * T) / sgma / np.sqrt(T)
    d2 = d1 - sgma * np.sqrt(T)
    return K * np.exp(-r * T) * N(-d2) - S * np.exp((b - r) * T) * N(-d1)



import numpy as np
from PyMPDATA import Solver, Stepper, ScalarField, VectorField
from PyMPDATA.boundary_conditions import Extrapolated, Constant
from PyMPDATA.impl.enumerations import INNER, OUTER
from .arakawa_c import arakawa_c


class MPDATA:
    # pylint: disable=too-few-public-methods
    def __init__(self, nz, dt, qv_of_zZ_at_t0, g_factor_of_zZ, nr, options, activation_bc):
        self.t = 0
        self.dt = dt
        self.fields = ('qv', 'ql')

        self.options = options

        self._solvers = {}
        for k in self.fields:
            grid = (nz, nr) if nr > 1 and k == 'ql' else (nz,)

            bcs_extrapol = tuple(
                Extrapolated(dim=d)
                for d in ((OUTER, INNER) if k == 'ql' and nr > 1 else (INNER,))
            )
            bcs_zero = tuple(
                Extrapolated(dim=d)
                for d in ((OUTER, INNER) if k == 'ql' and nr > 1 else (INNER,))
            )

            stepper = Stepper(options=self.options, n_dims=len(grid), non_unit_g_factor=True)

            data = g_factor_of_zZ(arakawa_c.z_scalar_coord(grid))
            if nr > 1 and k == 'ql':
                data = np.repeat(data.reshape(-1, 1), nr, axis=1).squeeze()
            g_factor = ScalarField(
                data=data,
                halo=self.options.n_halo,
                boundary_conditions=bcs_extrapol
            )

            if nr == 1 or k == 'qv':
                data = (np.zeros(nz + 1),)
            else:
                data = (np.zeros((nz + 1, nr)), np.zeros((nz, nr + 1)))
            advector = VectorField(
                data=data,
                halo=self.options.n_halo,
                boundary_conditions=bcs_zero
            )
            if k == 'qv':
                data = qv_of_zZ_at_t0(arakawa_c.z_scalar_coord(grid))
                bcs = (Constant(value=data[0]),)
            else:
                data = np.zeros(grid)
                if nr == 1:
                    bcs = (Constant(value=0),)
                else:
                    bcs = (
                        Constant(value=0),
                        activation_bc
                    )
            advectee = ScalarField(
                data=data,
                halo=self.options.n_halo,
                boundary_conditions=bcs
            )
            self._solvers[k] = Solver(
                stepper=stepper, advectee=advectee, advector=advector, g_factor=g_factor
            )

    def __getitem__(self, k):
        return self._solvers[k]



import numpy as np
from joblib import Parallel, delayed, parallel_backend
from PyMPDATA_examples.Arabas_and_Farhat_2020.simulation import Simulation
from PyMPDATA_examples.Arabas_and_Farhat_2020.setup2_american_put import Settings
from PyMPDATA_examples.Arabas_and_Farhat_2020.analysis_figures_2_and_3 import error_L2_norm


def compute_row(simulations):
    S0 = simulations[0].settings.S0
    T = simulations[0].settings.T
    for i in range(1, len(simulations)):
        assert simulations[i].settings.T == T
        assert simulations[i].settings.S0 == S0
    row = [T, S0]
    f = None
    for simulation in simulations:
        f = simulation.run(n_iters=2)
        row.append(
            error_L2_norm(
                simulation.solvers, simulation.settings, simulation.S, simulation.nt, n_iters=2
            )
        )
        np.testing.assert_almost_equal(simulation.S[simulation.ix_match], S0)
    row.append(f[simulations[-1].ix_match])
    row.append(simulations[0].settings.analytical_solution(S0))
    row.append(simulations[0].settings.analytical_solution(S0, amer=False))
    return row


def table_1_data():
    with parallel_backend('threading', n_jobs=-2):
        result = Parallel(verbose=10)(
            delayed(compute_row)(tuple(
                Simulation(Settings(T=T, C_opt=C_opt, S0=S0))
                for C_opt in (.02, .01, .005)
            ))
            for T in (.25, .5, 3)
            for S0 in (80, 90, 100, 110, 120)
        )
        return result



import os
import re
import sys
import pathlib
import pytest
from ghapi.all import GhApi, paged
from fastcore.net import ExceptionsHTTP


# https://stackoverflow.com/questions/7012921/recursive-grep-using-python
def findfiles(path, regex):
    reg_obj = re.compile(regex)
    res = []
    for root, _, fnames in os.walk(path):
        for fname in fnames:
            if reg_obj.match(fname):
                res.append(os.path.join(root, fname))
    return res


def grep(filepath, regex):
    reg_obj = re.compile(regex)
    res = []
    with open(filepath, encoding="utf8") as f:
        for line in f:
            if reg_obj.match(line):
                res.append(line)
    return res


@pytest.fixture(params=findfiles(
    pathlib.Path(__file__).parent.parent.absolute(),
    r'.*\.(ipynb|py|txt|yml|m|jl|md)$'
))
def file(request):
    return request.param


@pytest.fixture(scope='session')
def gh_issues():
    res = {}
    if 'CI' not in os.environ or ('GITHUB_ACTIONS' in os.environ and sys.version_info.minor >= 8):
        try:
            api = GhApi(owner='atmos-cloud-sim-uj', repo='PyMPDATA')
            pages = paged(
                api.issues.list_for_repo,
                owner='atmos-cloud-sim-uj',
                repo='PyMPDATA', state='all', per_page=100
            )
            for page in pages:
                for item in page.items:
                    res[item.number] = item.state
        except ExceptionsHTTP[403]:
            pass
    return res


def test_todos_annotated(file, gh_issues):  # pylint: disable=redefined-outer-name
    if (
        os.path.basename(file) == 'test_todos_annotated.py' or
        file.endswith("-checkpoint.ipynb") or
        ".eggs" in file
    ):
        return
    for line in grep(file, r'.*TODO.*'):
        if line.count('"image/png":') > 0:
            continue
        match = re.search(r'TODO #(\d+)', line)
        if match is None:
            raise Exception(f"TODO not annotated with issue id ({line})")
        giving_up_with_hope_other_builds_did_it = len(gh_issues) == 0
        if not giving_up_with_hope_other_builds_did_it:
            number = int(match.group(1))
            if number not in gh_issues.keys():
                raise Exception(f"TODO annotated with non-existent id ({line})")
            if gh_issues[number] != 'open':
                raise Exception(f"TODO remains for a non-open issue ({line})")



from collections import namedtuple
import numpy as np
from scipy import constants


def convert_to(value, unit):
    value /= unit


si = namedtuple(
    'si',
    ('kg', 'm', 's', 'metres', 'second', 'um', 'hPa', 'micrometre',
     'minutes', 'km', 'dimensionless', 'kelvin', 'mg')
)(
    kg=1.,
    m=1.,
    s=1.,
    metres=1.,
    second=1.,
    um=1e-6,
    hPa=100.,
    micrometre=1e-6,
    minutes=60.,
    km=1000.,
    dimensionless=1.,
    kelvin=1.,
    mg=1e-6
)

_Mv = 0.018015
_Md = 0.028970

const = namedtuple(
    'const',
    ('eps', 'g', 'p1000', 'Rd', 'Rv', 'c_pd', 'c_pv', 'lv',
     'rho_l', 'T0', 'ARM_C1', 'ARM_C2', 'ARM_C3')
)(
    eps=_Mv / _Md,
    g=constants.g,
    p1000=1000 * si.hPa,
    Rd=287.0027,
    Rv=constants.R / _Mv,
    c_pd=1005,
    c_pv=1850,
    lv=2.5e6,
    rho_l=1e3 * si.kg / si.m**3,
    T0=constants.zero_Celsius,
    ARM_C1=6.1094 * si.hPa,
    ARM_C2=17.625 * si.dimensionless,
    ARM_C3=243.04 * si.kelvin,
)


def rho_d(p, qv, theta_std):
    return p * (1 - 1 / (1 + const.eps / qv)) / (
                np.power(p / const.p1000, const.Rd / const.c_pd) * const.Rd * theta_std)


def drho_dz(g, p, T, qv, lv, dql_dz=0):
    Rq = const.Rv / (1 / qv + 1) + const.Rd / (1 + qv)
    cp = const.c_pv / (1 / qv + 1) + const.c_pd / (1 + qv)
    rho = p / Rq / T
    return (g / T * rho * (Rq / cp - 1) - p * lv / cp / T ** 2 * dql_dz) / Rq


# A14 in libcloudph++ 1.0 paper
def temperature(rhod, thd):
    return thd * np.power(
        rhod * thd / const.p1000 * const.Rd,
        const.Rd / const.c_pd / (1 - const.Rd / const.c_pd)
    )


# A15 in libcloudph++ 1.0 paper
def pressure(rhod, T, qv):
    return rhod * (1 + qv) * (const.Rv / (1 / qv + 1) + const.Rd / (1 + qv)) * T


def th_dry(th_std, qv):
    return th_std * np.power(1 + qv / const.eps, const.Rd / const.c_pd)


def pvs_Celsius(T):
    return const.ARM_C1 * np.exp((const.ARM_C2 * T) / (T + const.ARM_C3))


def pv(p, qv):
    return p * qv / (qv + const.eps)



from .settings import Settings, si, const
from .formulae import convert_to
from .plot import plot
from .arakawa_c import arakawa_c
from .droplet_activation import DropletActivation
from .mpdata import MPDATA



import numpy as np


class arakawa_c:
    @staticmethod
    def z_scalar_coord(grid):
        zZ = np.linspace(1 / 2, grid[0] - 1 / 2, grid[0])
        return zZ

    @staticmethod
    def z_vector_coord(grid):
        zZ = np.linspace(0, grid[0], grid[0] + 1)
        return zZ



from matplotlib import pyplot
import numpy as np
import matplotlib


def to_n_n(y, _, __):
    return y


def to_n_s(y, r1, r2):
    return y * (r2 ** 2 + r1 ** 2 + r1 * r2) * 4 / 3 * np.pi


def to_n_v(y, r1, r2):
    return y * (r2 + r1) * (r2 ** 2 + r1 ** 2) / 4 * 4 / 3 * np.pi


def from_n_n(n_n, _):
    return 1 * n_n


class Plotter:
    def __init__(self, settings, plots=('n', 'm')):
        self.settings = settings
        self.plots = plots
        self.cdfarg, self.dcdfarg = np.linspace(
            settings.r_min.magnitude,
            settings.r_max.magnitude,
            512, retstep=True
        )
        self.cdfarg *= settings.r_min.units
        self.dcdfarg *= settings.r_max.units

        matplotlib.rcParams.update({'font.size': 16})
        if len(plots) == 1:
            self.figsize = (10,6)
        else:
            self.figsize = (10,11)
        self.fig, self.axs = pyplot.subplots(len(plots), 1, figsize=self.figsize)
        if len(plots) == 1:
            self.axs = (self.axs,)
        self.fig.tight_layout(pad=5.0)
        self.style_dict = {}
        self.style_palette = ['-', '-', '-', '-', '-.']

        self.settings.si.setup_matplotlib()

        if 'n' in plots:
            self.axs[plots.index('n')].yaxis.set_units(
                1 / self.settings.si.micrometre / self.settings.si.centimetre ** 3)
        if 'm' in plots:
            self.axs[plots.index('m')].yaxis.set_units(
                1 / self.settings.si.micrometre)

        for i in range(len(plots)):
            self.axs[i].xaxis.set_units(self.settings.si.micrometre)
            self.axs[i].grid()
            self.axs[i].set_xlabel('particle radius [$μm$]')

        if 'm' in plots:
            self.axs[plots.index('m')].set_ylabel('(dM/dr)/M [$μm^{-1}$]')
        if 'n' in plots:
            self.axs[plots.index('n')].set_ylabel('dN/dr [$cm^{-3} μm^{-1}$]')

    def pdf_curve(self, pdf, mnorm, color='black'):
        x = self.cdfarg
        y = pdf(x)

        # normalised mass distribution
        if 'm' in self.plots:
            y_mass = y * x ** 3 * 4 / 3 * np.pi * self.settings.rho_w / self.settings.rho_a / mnorm
            self.axs[self.plots.index('m')].plot(x, y_mass, color=color, linestyle=':')

        # number distribution
        if 'n' in self.plots:
            self.axs[self.plots.index('n')].plot(x, y, color=color, linestyle=':')

    def _plot(self, index, x, y, fill, lbl, color, linewidth):
        if fill:
            self.axs[self.plots.index(index)].fill_between(
                x,
                y,
                step='mid', label=lbl, color='gainsboro', alpha = 1)
        else:
            self.axs[self.plots.index(index)].step(
                x,
                y,
                where='mid', label=lbl, color=color, linewidth=linewidth
            )
        self.xlim(index)

    def pdf_histogram(self, x, y,
                      bin_boundaries, label, mnorm, color='black', linewidth = 1, fill=True):
        r1 = bin_boundaries[:-1]
        r2 = bin_boundaries[1:]

        lbl = label
        if label not in self.style_dict:
            self.style_dict[label] = self.style_palette[len(self.style_dict)]
        else:
            lbl = ''

        # number distribution
        if 'n' in self.plots:
            self._plot('n', x, to_n_n(y, r1, r2),
                       fill=fill, lbl=lbl, color=color, linewidth=linewidth)

        # normalised mass distribution
        if 'm' in self.plots:
            self._plot('m', x,
                       to_n_v(y, r1, r2) * self.settings.rho_w / self.settings.rho_a / mnorm,
                       fill=fill, color=color, linewidth=linewidth, lbl=lbl)

    def xlim(self, plot):
        self.axs[self.plots.index(plot)].set_xlim(
            (0, self.settings.r_max.magnitude)
        )



from pkg_resources import get_distribution, DistributionNotFound, VersionConflict
try:
    __version__ = get_distribution(__name__).version
except (DistributionNotFound, VersionConflict):
    # package is not installed
    pass



# adapted from:
# https://github.com/igfuw/shallow-water-elliptic-drop/blob/master/analytical/analytic_equations.py
# Code used in the paper of Jarecka, Jaruga, Smolarkiewicz -
# "A Spreading Drop of Shallow Water" (JCP 289, doi:10.1016/j.jcp.2015.02.003).

import numpy as np
import numba
from scipy.integrate import odeint


def amplitude(x, y, lx, ly):
    A = 1 / lx / ly
    h = A * (1 - (x / lx) ** 2 - (y / ly) ** 2)
    return np.where(h > 0, h, 0)

@numba.njit()
def deriv(y, _):
    """
    return derivatives of [lambda_x, dlambda_x/dt, lambda_y, dlambda_y/dt
    four first-order ODEs based on  eq. 7  (Jarecka, Jaruga, Smolarkiewicz)
    """
    return np.array((y[1], 2. / y[0] ** 2 / y[2], y[3], 2. / y[0] / y[2] ** 2))


def d2_el_lamb_lamb_t_evol(times, lamb_x0, lamb_y0):
    """
    solving coupled nonlinear second-order ODEs - eq. 7  (Jarecka, Jaruga, Smolarkiewicz)
    returning array with first dim denoting time, second dim:
    [lambda_x, dot{lambda_x}, lambda_y, dot{lambda_y}
    """
    assert times[0] == 0
    yinit = np.array([lamb_x0, 0., lamb_y0, 0.])  # initial values (dot_lamb = 0.)
    result, info = odeint(deriv, yinit, times, full_output=True)
    assert info['message'] == 'Integration successful.'
    return result



""" the magick behind ``pip install ...`` """
from setuptools import setup, find_packages


def get_long_description():
    """ returns contents of README.md file """
    with open("README.md", "r", encoding="utf8") as file:
        long_description = file.read()
    return long_description


setup(
    name='PyMPDATA-examples',
    description='PyMPDATA usage examples reproducing results from literature'
                ' and depicting how to use PyMPDATA in Python from Jupyter notebooks',
    use_scm_version={"local_scheme": lambda _: "", "version_scheme": "post-release"},
    setup_requires=['setuptools_scm'],
    install_requires=['PyMPDATA',
                      'atmos_cloud_sim_uj_utils',
                      'pystrict',
                      'matplotlib',
                      'ipywidgets',
                      'scipy',
                      'pint',
                      'joblib',
                      'sympy',
                      'ghapi'],
    author='https://github.com/orgs/atmos-cloud-sim-uj/people',
    license="GPL-3.0",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(include=['PyMPDATA_examples', 'PyMPDATA_examples.*']),
    package_data={'': ['*/*/*.txt']},
    include_package_data=True,
    project_urls={
        "Tracker": "https://github.com/atmos-cloud-sim-uj/PyMPDATA/issues",
        "Documentation": "https://atmos-cloud-sim-uj.github.io/PyMPDATA-examples",
        "Source": "https://github.com/atmos-cloud-sim-uj/PyMPDATA-examples"
    }
)



import numpy as np
from PyMPDATA import VectorField
from PyMPDATA.boundary_conditions import Periodic


def nondivergent_vector_field_2d(grid, size, dt, stream_function: callable, halo):
    dx = size[0] / grid[0]
    dz = size[1] / grid[1]
    dxX = 1 / grid[0]
    dzZ = 1 / grid[1]

    xX, zZ = x_vec_coord(grid)
    rho_velocity_x = -(stream_function(xX, zZ + dzZ/2) - stream_function(xX, zZ - dzZ/2)) / dz

    xX, zZ = z_vec_coord(grid)
    rho_velocity_z = (stream_function(xX + dxX/2, zZ) - stream_function(xX - dxX/2, zZ)) / dx

    GC = [rho_velocity_x * dt / dx, rho_velocity_z * dt / dz]

    # CFL condition
    for val in GC:
        np.testing.assert_array_less(np.abs(val), 1)

    result = VectorField(GC, halo=halo, boundary_conditions=(Periodic(), Periodic()))

    # nondivergence (of velocity field, hence dt)
    assert np.amax(abs(result.div((dt, dt)).get())) < 5e-9

    return result


def x_vec_coord(grid):
    nx = grid[0]+1
    nz = grid[1]
    xX = np.repeat(np.linspace(0, grid[0], nx).reshape((nx, 1)), nz, axis=1) / grid[0]
    assert np.amin(xX) == 0
    assert np.amax(xX) == 1
    assert xX.shape == (nx, nz)
    zZ = np.repeat(np.linspace(1 / 2, grid[1] - 1/2, nz).reshape((1, nz)), nx, axis=0) / grid[1]
    assert np.amin(zZ) >= 0
    assert np.amax(zZ) <= 1
    assert zZ.shape == (nx, nz)
    return xX, zZ


def z_vec_coord(grid):
    nx = grid[0]
    nz = grid[1]+1
    xX = np.repeat(np.linspace(1/2, grid[0]-1/2, nx).reshape((nx, 1)), nz, axis=1) / grid[0]
    assert np.amin(xX) >= 0
    assert np.amax(xX) <= 1
    assert xX.shape == (nx, nz)
    zZ = np.repeat(np.linspace(0, grid[1], nz).reshape((1, nz)), nx, axis=0) / grid[1]
    assert np.amin(zZ) == 0
    assert np.amax(zZ) == 1
    assert zZ.shape == (nx, nz)
    return xX, zZ



""" example based on the solid-body rotation test from
 [Molenkap 1968](http://dx.doi.org/10.1175/1520-0450(1968)007%3C0160:AOFDMA%3E2.0.CO;2)
 (as in Fig. 12 in [Jaruga et al. 2015](https://doi.org/10.5194/gmd-8-1005-2015)) """



import numpy as np
from scipy import integrate


def from_pdf_2d(pdf: callable, xrange: list, yrange: list, gridsize: list):
    z = np.empty(gridsize)
    dx, dy = (xrange[1] - xrange[0]) / gridsize[0], (yrange[1] - yrange[0]) / gridsize[1]
    for i in range(gridsize[0]):
        for j in range(gridsize[1]):
            z[i, j] = integrate.nquad(pdf, ranges=(
                (xrange[0] + dx*i, xrange[0] + dx*(i+1)),
                (yrange[0] + dy*j, yrange[0] + dy*(j+1))
            ))[0] / dx / dy
    x = np.linspace(xrange[0] + dx / 2, xrange[1] - dx / 2, gridsize[0])
    y = np.linspace(yrange[0] + dy / 2, yrange[1] - dy / 2, gridsize[1])
    return x, y, z


def from_cdf_1d(cdf: callable, x_min: float, x_max: float, nx: int):
    dx = (x_max - x_min) / nx
    x = np.linspace(x_min + dx / 2, x_max - dx / 2, nx)
    xh = np.linspace(x_min, x_max, nx + 1)
    y = np.diff(cdf(xh)) / dx
    return x, y


def discretised_analytical_solution(rh, pdf_t, midpoint_value=False, r=None):
    if midpoint_value:
        assert r is not None
    else:
        assert r is None
    output = np.empty(rh.shape[0]-1)
    for i in range(output.shape[0]):
        if midpoint_value:
            output[i] = pdf_t(r[i])
        else:
            dcdf, _ = integrate.quad(pdf_t, rh[i], rh[i + 1])
            output[i] = dcdf / (rh[i + 1] - rh[i])
    return output



from PyMPDATA import Options, Stepper, ScalarField, Solver
from PyMPDATA.boundary_conditions import Periodic
from PyMPDATA_examples.utils.discretisation import from_pdf_2d
from PyMPDATA_examples.utils import nondivergent_vector_field_2d
from PyMPDATA_examples.Molenkamp_test_as_in_Jaruga_et_al_2015_Fig_12.settings import Settings


class Simulation:
    def __init__(self, settings: Settings, options: Options):
        _, __, z = from_pdf_2d(
            settings.pdf,
            xrange=settings.xrange,
            yrange=settings.yrange,
            gridsize=settings.grid
        )
        stepper = Stepper(options=options, grid=settings.grid, non_unit_g_factor=False)
        advector = nondivergent_vector_field_2d(
            settings.grid,
            settings.size,
            settings.dt,
            settings.stream_function,
            options.n_halo
        )
        advectee = ScalarField(z.astype(dtype=options.dtype), halo=options.n_halo,
                               boundary_conditions=(Periodic(), Periodic()))
        self.mpdata = Solver(stepper=stepper, advectee=advectee, advector=advector)
        self.nt = settings.nt

    @property
    def state(self):
        return self.mpdata.advectee.get().copy()

    def run(self):
        self.mpdata.advance(self.nt)



import numpy as np


class DrDt:
    """ eq. 7.20 in Rogers and Yau 1989 """
    def __init__(self, ksi_1, S):
        self.ksi = (S - 1) * ksi_1

    def __call__(self, r):
        return self.ksi / r

    def mean(self, r1, r2):
        return self.ksi * np.log(r2/r1) / (r2 - r1)


# pylint: disable=too-few-public-methods
class PdfEvolver:
    """ eq. 7.32 in Rogers and Yau 1989 """
    def __init__(self, pdf, drdt: DrDt, t):
        self.t = t
        self.pdf = pdf
        self.drdt = drdt

    def __call__(self, r):
        with np.errstate(invalid='ignore'):
            arg = np.sqrt(r ** 2 - 2 * self.drdt.ksi * self.t)
        result = r / arg * self.pdf(arg)

        if isinstance(result.magnitude, np.ndarray):
            result = np.where(np.isfinite(result.magnitude), result.magnitude, 0) * result.units
        else:
            if not np.isfinite(result):
                result = 0 * result.units

        return result



import pathlib
import numpy as np
from PyMPDATA import Options
from PyMPDATA_examples.Olesik_et_al_2022.settings import (
    Settings, default_nr, default_GC_max, default_opt_set
)
from PyMPDATA_examples.Olesik_et_al_2022.coordinates import x_id, x_log_of_pn
from PyMPDATA_examples.Olesik_et_al_2022.simulation import Simulation


grid_layout_set = (x_log_of_pn(r0=1, base=2),)
opt_set = default_opt_set.values()

RTOL = 1.5


def test_wall_time(n_runs=3, mrats=(10,), generate=False, print_tab=True, rtol=RTOL):
    settings = Settings(nr=default_nr * 10, mixing_ratios_g_kg=np.array(mrats))
    table_data = {"opts": [], "values": []}
    for grid in grid_layout_set:
        for opts in opt_set:
            i = 0
            minimum_values = []
            while i < n_runs:
                result = make_data(settings, grid, opts)
                wall_times = result['wall_time']
                minimal = np.nanmin(wall_times)
                minimum_values.append(minimal)
                i += 1
            selected_value = np.min(minimum_values)
            if opts == {'n_iters': 1}:
                norm = selected_value
            table_data["opts"].append(str(opts) + "(" + grid.__class__.__name__ + ")")
            table_data["values"].append(round(selected_value / norm, 1))
    make_textable(data=table_data, generate=generate, print_tab=print_tab)
    compare_refdata(data=table_data, rtol=rtol, generate=generate)


def make_data(settings, grid, opts):
    options = Options(**opts)
    simulation = Simulation(
        settings=settings, grid_layout=grid, psi_coord=x_id(),
        opts=options, GC_max=default_GC_max)
    result = {"wall_time": []}
    last_step = 0
    for n_steps in simulation.out_steps:
        steps = n_steps - last_step
        wall_time_per_timestep = simulation.step(steps)
        last_step += steps
        result['wall_time'].append(wall_time_per_timestep)
    return result


def make_textable(data, generate=False, print_tab=False):
    latex_data = r"\hline" + " Variant  & Elapsed Real Time (wrt upwind) " + r"\\ \hline" + "\n"
    for opt, value in zip(data["opts"], data["values"]):
        latex_data += r"\hline" + f" {opt} & {value} " + r"\\ \hline" + "\n"
    latex_start = r"\begin{table}[]" + "\n" + r"\begin{tabular}{| l | l |}" + "\n"
    latex_end = r"\end{tabular}" + "\n" + r"\end{table}"
    latex_table = latex_start + latex_data + latex_end
    if print_tab:
        print(latex_table)
    with open(
        pathlib.Path(__file__).parent.joinpath("wall_time_textable.txt"),
        "w+" if generate else "r",
        encoding='utf-8'
    ) as f:
        if generate:
            f.write(latex_table)


def compare_refdata(data, rtol, generate=False):
    delimiter = ';'
    path = pathlib.Path(__file__).parent.joinpath("wall_time_refdata.txt")
    if generate:
        table = np.char.array(
            np.concatenate([data['opts'], data['values']])
        ).reshape(2, len(data['values'])).T
        np.savetxt(path, table, delimiter=delimiter, fmt="%s")
    else:
        table = np.loadtxt(path, delimiter=delimiter, dtype=str)
        np.testing.assert_allclose(
            actual=data['values'],
            desired=np.array(table[:,1].astype(float)), rtol=rtol)
        np.testing.assert_array_equal(data['opts'], table[:,0])



import numpy as np
from pystrict import strict
import PyMPDATA_examples.Arabas_and_Farhat_2020.Black_Scholes_1973 as BS73


@strict
class Settings:
    S0 = 55
    T = 0.5
    amer = False
    S_min = 10
    S_max = 2000
    sigma = 0.6
    r = 0.008
    K1 = 75
    K2 = 175
    S_match = 175

    def __init__(self, *, n_iters: int = 2, l2_opt: int = 2, C_opt: float = 0.034):
        self.n_iters = n_iters
        self.l2_opt = l2_opt
        self.C_opt = C_opt

    def payoff(self, S: np.ndarray):
        return np.exp(-self.r * self.T) * (np.maximum(0, self.K2 - S) - np.maximum(0, self.K1 - S))

    def analytical_solution(self, S: np.ndarray):
        return (
                BS73.p_euro(S, K=self.K2, T=self.T, r=self.r, b=self.r, sgma=self.sigma) -
                BS73.p_euro(S, K=self.K1, T=self.T, r=self.r, b=self.r, sgma=self.sigma)
        )



from typing import Optional
from pystrict import strict
from scipy.interpolate import interp1d
from scipy.integrate import solve_ivp
import numpy as np
from PyMPDATA_examples.Olesik_et_al_2022.settings import ksi_1 as default_ksi_1
from . import formulae
from .formulae import si, const
from .arakawa_c import arakawa_c


@strict
class Settings:
    def __init__(self, dt: float, dz: float, rhod_w_const: float,
                 t_max: float = 15 * si.minutes, nr: int = 1,
                 r_min: float = np.nan, r_max: float = np.nan, p0: Optional[float] = None,
                 ksi_1: float = default_ksi_1.to_base_units().magnitude,
                 z_max: float = 3000 * si.metres):
        self.dt = dt
        self.dz = dz

        self.nr = nr
        self.ksi_1 = ksi_1

        self.z_max = z_max
        self.t_max = t_max

        self.qv = interp1d((0, 740, 3260), (.015, .0138, .0024))
        self._th = interp1d((0, 740, 3260), (297.9, 297.9, 312.66))

        # note: not in the paper,
        # https://github.com/BShipway/KiD/tree/master/src/physconst.f90#L43
        p0 = p0 or 1000 * si.hPa

        self.rhod0 = formulae.rho_d(p0, self.qv(0), self._th(0))
        self.thd = lambda z: formulae.th_dry(self._th(z), self.qv(z))

        def drhod_dz(z, rhod):
            T = formulae.temperature(rhod[0], self.thd(z))
            p = formulae.pressure(rhod[0], T, self.qv(z))
            return formulae.drho_dz(const.g, p, T, self.qv(z), const.lv)

        z_points = np.arange(0, self.z_max + self.dz / 2, self.dz / 2)
        rhod_solution = solve_ivp(
            fun=drhod_dz,
            t_span=(0, self.z_max),
            y0=np.asarray((self.rhod0,)),
            t_eval=z_points
        )
        assert rhod_solution.success

        self.rhod = interp1d(z_points, rhod_solution.y[0])

        self.t_1 = 600 * si.s
        self.rhod_w = lambda t: rhod_w_const * np.sin(np.pi * t / self.t_1) if t < self.t_1 else 0

        self.r_min = r_min
        self.r_max = r_max
        self.bin_boundaries, self.dr = np.linspace(
            self.r_min,
            self.r_max,
            self.nr + 1, retstep=True
        )

        self.dr_power = {}
        for k in (1, 2, 3, 4):
            self.dr_power[k] = self.bin_boundaries[1:] ** k - self.bin_boundaries[:-1] ** k
            self.dr_power[k] = self.dr_power[k].reshape(1, -1).T

        self.z_vec = self.dz * arakawa_c.z_vector_coord((self.nz,))

    @property
    def nz(self):
        nz = self.z_max / self.dz
        assert nz == int(nz)
        return int(nz)

    @property
    def nt(self):
        nt = self.t_max / self.dt
        assert nt == int(nt)
        return int(nt)



import numpy as np
from joblib import Parallel, delayed, parallel_backend
from PyMPDATA_examples.Arabas_and_Farhat_2020.simulation import Simulation
from PyMPDATA_examples.Arabas_and_Farhat_2020.setup1_european_corridor import Settings
from PyMPDATA_examples.utils.error_norms import L2



def compute(simulation):
    output = []
    for n_iters in (1, 2):
        simulation.run(n_iters)
        output.append({
            "n_iters": n_iters,
            "log2_C": np.log2(simulation.C),
            "log2_C_opt": np.log2(simulation.settings.C_opt),
            "log2_l2": np.log2(simulation.l2),
            "log2_l2_opt": np.log2(simulation.settings.l2_opt),
            "err2": error_L2_norm(
                simulation.solvers, simulation.settings, simulation.S, simulation.nt, n_iters)
        })
    return output


def convergence_in_space(num=8):
    with parallel_backend('threading', n_jobs=-2):
        data = Parallel(verbose=10)(
            delayed(compute)(Simulation(Settings(l2_opt=2 ** log2_l2_opt, C_opt=2 ** log2_C_opt)))
            for log2_C_opt in np.linspace(-9.5, -6, num=num)
            for log2_l2_opt in range(1, 4)
        )
        result = {}
        for pair in data:
            for datum in pair:
                label = f" $\\lambda^2\\approx2^{{{datum['log2_l2_opt']}}}$"
                key = ("upwind" + label, "MPDATA" + label)[datum["n_iters"]-1]
                if key not in result:
                    result[key] = ([], [])
                result[key][0].append(datum["log2_C"])
                result[key][1].append(datum["err2"])
        return result


def convergence_in_time(num=13):
    with parallel_backend('threading', n_jobs=-2):
        data = Parallel(verbose=10)(
            delayed(compute)(Simulation(Settings(l2_opt=2 ** log2_l2_opt, C_opt=2 ** log2_C_opt)))
            for log2_C_opt in np.log2((.01, .005, .0025))
            for log2_l2_opt in np.linspace(1.1, 3.5, num=num)
        )
        result = {}
        for pair in data:
            for datum in pair:
                label = f" $C\\approx{2**(datum['log2_C_opt']):.4f}$"
                key = ("upwind" + label, "MPDATA" + label)[datum['n_iters']-1]
                if key not in result:
                    result[key] = ([], [])
                result[key][0].append(datum['log2_l2'])
                result[key][1].append(datum['err2'])
        return result


def error_L2_norm(solvers, settings, S, nt, n_iters: int):
    numerical = solvers[n_iters].advectee.get()
    analytical = settings.analytical_solution(S)
    return L2(numerical, analytical, nt)



import numpy as np
from matplotlib import pyplot, pylab
import matplotlib
from .formulae import convert_to, si


def plot(var, mult, label, output,
         rng=None, threshold=None, cmap='copper', rasterized=False, figsize=None):
    lines = {3: ':', 6: '--', 9: '-', 12: '-.'}
    colors = {3: 'crimson', 6: 'orange', 9: 'navy', 12: 'green'}
    fctr = 50  # rebin by fctr in time dimension (https://gist.github.com/zonca/1348792)

    dt = (output['t'][1] - output['t'][0]) * fctr
    dz = output['z'][1] - output['z'][0]
    tgrid = np.concatenate(((output['t'][0] - dt / 2,), output['t'][0::fctr] + dt / 2))
    zgrid = np.concatenate(((output['z'][0] - dz / 2,), output['z'] + dz / 2))
    convert_to(zgrid, si.km)

    fig = pyplot.figure(constrained_layout=True, figsize=figsize)
    gs = fig.add_gridspec(25, 5)
    ax1 = fig.add_subplot(gs[:-1, 0:3])

    data = output[var]*mult
    lngth = data.shape[1] - 1
    assert (lngth//fctr)*fctr == lngth

    tmp = np.empty((data.shape[0], (data.shape[1]-1)//fctr+1))
    tmp[:, 0] = data[:, 0]
    M, N = data[:, 1:].shape
    m, n = M, N//fctr
    tmp[:, 1:] = data[:, 1:].reshape((m, M//m, n, N//n)).mean(3).mean(1)
    data = tmp

    if threshold is not None:
        data[data < threshold] = np.nan
    mesh = ax1.pcolormesh(tgrid / si.minutes, zgrid,
                          data,
                          cmap=cmap, rasterized=rasterized,
                          vmin=None if rng is None else rng[0],
                          vmax=None if rng is None else rng[1]
                          )

    ax1.set_xlabel('time [min]')
    ax1.set_xticks(list(lines.keys()))
    ax1.set_ylabel('z [km]')
    ax1.grid()

    cbar = fig.colorbar(mesh, fraction=.05, location='top')
    cbar.set_label(label)

    ax2 = fig.add_subplot(gs[:-1, 3:], sharey=ax1)
    ax2.set_xlabel(label)
    ax2.grid()
    if rng is not None:
        ax2.set_xlim(rng)

    last_t = -1
    for i, t in enumerate(output['t']):
        x, z = output[var][:, i]*mult, output['z'].copy()
        convert_to(z, si.km)
        params = {'color': 'black'}
        for line_t, line_s in lines.items():
            if last_t < line_t * si.minutes <= t:
                params['ls'] = line_s
                params['color'] = colors[line_t]
                ax2.step(x, z-(dz/si.km)/2, where='pre', **params)
                ax1.axvline(t/si.minutes, **params)
        last_t = t


def plot_3d(psi, settings, options, r_min, r_max, max_height):
    max_psi = np.amax(psi / (1 / si.mg / si.um))
    if max_psi > max_height:
        raise ValueError(f"max(psi)={max_psi} > {max_height}")
    pylab.figure(figsize=(10, 5))
    ax = pylab.subplot(projection='3d')
    ax.view_init(75, 85)

    cmap = matplotlib.cm.get_cmap('gray_r')
    min_height = 0

    dz = np.rot90(psi, 2).flatten()
    rgba = [cmap((k - min_height) / (1 + max_height)) for k in dz]
    dz[dz < .05 * max_height] = np.nan
    factor = .9
    ax.bar3d(
        *[arr.flatten() for arr in np.meshgrid(
            (settings.bin_boundaries[-2::-1] + settings.dr * (1 - factor) / 2) / si.um,
            (np.arange(settings.nz - 1, -1, -1) + (1 - factor) / 2) * settings.dz / si.km
        )],
        z=0,
        dx=factor * settings.dr / si.um,
        dy=factor * settings.dz / si.km,
        dz=dz / (1 / si.mg / si.um),
        shade=True,
        color=rgba,
        lightsource=matplotlib.colors.LightSource(azdeg=-64, altdeg=15),
        zsort='max'
    )
    ax.set_title(f"MPDATA iterations: {options.n_iters}")
    ax.set_xlabel("droplet radius [μm]")
    ax.set_ylabel("z [km]")
    ax.set_zlabel("psi [1/μm 1/mg 1/m]")
    ax.set_zlim(0, max_height / (1 / si.mg / si.um))
    ax.set_zticks([0, max_height / (1 / si.mg / si.um)])
    ax.set_ylim(settings.z_max / si.km, 0)
    ax.set_xlim(r_max / si.um, r_min / si.um)
    nticks = 6
    ax.set_xticks((r_min + np.arange(nticks + 1) * (r_max - r_min) / nticks) / si.um)
